# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import Levenshtein
import string


class RecMetric(object):
    def __init__(self, main_indicator='acc', is_filter=False, **kwargs):
        self.main_indicator = main_indicator
        self.is_filter = is_filter
        self.reset()

    def _normalize_text(self, text):
        text = ''.join(
            filter(lambda x: x in (string.digits + string.ascii_letters), text))
        return text.lower()

    def __call__(self, pred_label, *args, **kwargs):
        preds, labels = pred_label
        correct_num = 0
        all_num = 0
        norm_edit_dis = 0.0
        for (pred, pred_conf), (target, _) in zip(preds, labels):
            pred = pred.replace(" ", "")
            target = target.replace(" ", "")
            if self.is_filter:
                pred = self._normalize_text(pred)
                target = self._normalize_text(target)
            norm_edit_dis += Levenshtein.distance(pred, target) / max(
                len(pred), len(target), 1)
            if pred == target:
                correct_num += 1
            all_num += 1
        self.correct_num += correct_num
        self.all_num += all_num
        self.norm_edit_dis += norm_edit_dis
        return {
            'acc': correct_num / all_num,
            'norm_edit_dis': 1 - norm_edit_dis / (all_num + 1e-3)
        }

    def get_metric(self):
        """
        return metrics {
                 'acc': 0,
                 'norm_edit_dis': 0,
            }
        """
        acc = 1.0 * self.correct_num / (self.all_num + 1e-3)
        norm_edit_dis = 1 - self.norm_edit_dis / (self.all_num + 1e-3)
        self.reset()
        return {'acc': acc, 'norm_edit_dis': norm_edit_dis}

    def reset(self):
        self.correct_num = 0
        self.all_num = 0
        self.norm_edit_dis = 0

    @classmethod
    def eval(cls, preds, labels):
        preds = [(x, 1) for x in preds]
        labels = [(x, 1) for x in labels]
        return cls()([preds, labels])



"""
# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys
import six
import cv2
import numpy as np


class GenTableMask(object):
    """ gen table mask """

    def __init__(self, shrink_h_max, shrink_w_max, mask_type=0, **kwargs):
        self.shrink_h_max = 5
        self.shrink_w_max = 5
        self.mask_type = mask_type
        
    def projection(self, erosion, h, w, spilt_threshold=0):
        # 水平投影
        projection_map = np.ones_like(erosion)
        project_val_array = [0 for _ in range(0, h)]

        for j in range(0, h):
            for i in range(0, w):
                if erosion[j, i] == 255:
                    project_val_array[j] += 1
        # 根据数组，获取切割点
        start_idx = 0  # 记录进入字符区的索引
        end_idx = 0  # 记录进入空白区域的索引
        in_text = False  # 是否遍历到了字符区内
        box_list = []
        for i in range(len(project_val_array)):
            if in_text == False and project_val_array[i] > spilt_threshold:  # 进入字符区了
                in_text = True
                start_idx = i
            elif project_val_array[i] <= spilt_threshold and in_text == True:  # 进入空白区了
                end_idx = i
                in_text = False
                if end_idx - start_idx <= 2:
                    continue
                box_list.append((start_idx, end_idx + 1))

        if in_text:
            box_list.append((start_idx, h - 1))
        # 绘制投影直方图
        for j in range(0, h):
            for i in range(0, project_val_array[j]):
                projection_map[j, i] = 0
        return box_list, projection_map

    def projection_cx(self, box_img):
        box_gray_img = cv2.cvtColor(box_img, cv2.COLOR_BGR2GRAY)
        h, w = box_gray_img.shape
        # 灰度图片进行二值化处理
        ret, thresh1 = cv2.threshold(box_gray_img, 200, 255, cv2.THRESH_BINARY_INV)
        # 纵向腐蚀
        if h < w:
            kernel = np.ones((2, 1), np.uint8)
            erode = cv2.erode(thresh1, kernel, iterations=1)
        else:
            erode = thresh1
        # 水平膨胀
        kernel = np.ones((1, 5), np.uint8)
        erosion = cv2.dilate(erode, kernel, iterations=1)
        # 水平投影
        projection_map = np.ones_like(erosion)
        project_val_array = [0 for _ in range(0, h)]

        for j in range(0, h):
            for i in range(0, w):
                if erosion[j, i] == 255:
                    project_val_array[j] += 1
        # 根据数组，获取切割点
        start_idx = 0  # 记录进入字符区的索引
        end_idx = 0  # 记录进入空白区域的索引
        in_text = False  # 是否遍历到了字符区内
        box_list = []
        spilt_threshold = 0
        for i in range(len(project_val_array)):
            if in_text == False and project_val_array[i] > spilt_threshold:  # 进入字符区了
                in_text = True
                start_idx = i
            elif project_val_array[i] <= spilt_threshold and in_text == True:  # 进入空白区了
                end_idx = i
                in_text = False
                if end_idx - start_idx <= 2:
                    continue
                box_list.append((start_idx, end_idx + 1))

        if in_text:
            box_list.append((start_idx, h - 1))
        # 绘制投影直方图
        for j in range(0, h):
            for i in range(0, project_val_array[j]):
                projection_map[j, i] = 0
        split_bbox_list = []
        if len(box_list) > 1:
            for i, (h_start, h_end) in enumerate(box_list):
                if i == 0:
                    h_start = 0
                if i == len(box_list):
                    h_end = h
                word_img = erosion[h_start:h_end + 1, :]
                word_h, word_w = word_img.shape
                w_split_list, w_projection_map = self.projection(word_img.T, word_w, word_h)
                w_start, w_end = w_split_list[0][0], w_split_list[-1][1]
                if h_start > 0:
                    h_start -= 1
                h_end += 1
                word_img = box_img[h_start:h_end + 1:, w_start:w_end + 1, :]
                split_bbox_list.append([w_start, h_start, w_end, h_end])
        else:
            split_bbox_list.append([0, 0, w, h])
        return split_bbox_list

    def shrink_bbox(self, bbox):
        left, top, right, bottom = bbox
        sh_h = min(max(int((bottom - top) * 0.1), 1), self.shrink_h_max)
        sh_w = min(max(int((right - left) * 0.1), 1), self.shrink_w_max)
        left_new = left + sh_w
        right_new = right - sh_w
        top_new = top + sh_h
        bottom_new = bottom - sh_h
        if left_new >= right_new:
            left_new = left
            right_new = right
        if top_new >= bottom_new:
            top_new = top
            bottom_new = bottom
        return [left_new, top_new, right_new, bottom_new]

    def __call__(self, data):
        img = data['image']
        cells = data['cells']
        height, width = img.shape[0:2]
        if self.mask_type == 1:
            mask_img = np.zeros((height, width), dtype=np.float32)
        else:
            mask_img = np.zeros((height, width, 3), dtype=np.float32)
        cell_num = len(cells)
        for cno in range(cell_num):
            if "bbox" in cells[cno]:
                bbox = cells[cno]['bbox']
                left, top, right, bottom = bbox
                box_img = img[top:bottom, left:right, :].copy()
                split_bbox_list = self.projection_cx(box_img)
                for sno in range(len(split_bbox_list)):
                    split_bbox_list[sno][0] += left
                    split_bbox_list[sno][1] += top
                    split_bbox_list[sno][2] += left
                    split_bbox_list[sno][3] += top

                for sno in range(len(split_bbox_list)):
                    left, top, right, bottom = split_bbox_list[sno]
                    left, top, right, bottom = self.shrink_bbox([left, top, right, bottom])
                    if self.mask_type == 1:
                        mask_img[top:bottom, left:right] = 1.0
                        data['mask_img'] = mask_img
                    else:
                        mask_img[top:bottom, left:right, :] = (255, 255, 255)        
                        data['image'] = mask_img
        return data

class ResizeTableImage(object):
    def __init__(self, max_len, **kwargs):
        super(ResizeTableImage, self).__init__()
        self.max_len = max_len

    def get_img_bbox(self, cells):
        bbox_list = []
        if len(cells) == 0:
            return bbox_list
        cell_num = len(cells)
        for cno in range(cell_num):
            if "bbox" in cells[cno]:
                bbox = cells[cno]['bbox']
                bbox_list.append(bbox)
        return bbox_list

    def resize_img_table(self, img, bbox_list, max_len):
        height, width = img.shape[0:2]
        ratio = max_len / (max(height, width) * 1.0)
        resize_h = int(height * ratio)
        resize_w = int(width * ratio)
        img_new = cv2.resize(img, (resize_w, resize_h))
        bbox_list_new = []
        for bno in range(len(bbox_list)):
            left, top, right, bottom = bbox_list[bno].copy()
            left = int(left * ratio)
            top = int(top * ratio)
            right = int(right * ratio)
            bottom = int(bottom * ratio)
            bbox_list_new.append([left, top, right, bottom])
        return img_new, bbox_list_new
    
    def __call__(self, data):
        img = data['image']
        if 'cells' not in data:
            cells = []
        else:
            cells = data['cells']
        bbox_list = self.get_img_bbox(cells)
        img_new, bbox_list_new = self.resize_img_table(img, bbox_list, self.max_len)
        data['image'] = img_new
        cell_num = len(cells)
        bno = 0
        for cno in range(cell_num):
            if "bbox" in data['cells'][cno]:
                data['cells'][cno]['bbox'] = bbox_list_new[bno]
                bno += 1
        data['max_len'] = self.max_len
        return data

class PaddingTableImage(object):
    def __init__(self, **kwargs):
        super(PaddingTableImage, self).__init__()
    
    def __call__(self, data):
        img = data['image']
        max_len = data['max_len']
        padding_img = np.zeros((max_len, max_len, 3), dtype=np.float32)
        height, width = img.shape[0:2]
        padding_img[0:height, 0:width, :] = img.copy()
        data['image'] = padding_img
        return data
            


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#File: TL2p_icdar_1_1.py
#Version: 1.1
#Version info: changes for Python 3
#Date: 2019-12-29
#Description: Evaluation script that computes Text Localization following the algorithm of Wolf et al [1]
#1. C. Wolf and J.M. Jolion, "Object Count / Area Graphs for the Evaluation of Object Detection and Segmentation Algorithms", International Journal of Document Analysis, vol. 8, no. 4, pp. 280-296, 2006.

from collections import namedtuple
import pyxlpr.data.icdar.rrc_evaluation_funcs_1_1 as rrc_evaluation_funcs
import importlib

def evaluation_imports():
    """
    evaluation_imports: Dictionary ( key = module name , value = alias  )  with python modules used in the evaluation. 
    """    
    return {
            'math':'math',
            'numpy':'np'
            }

def default_evaluation_params():
    """
    default_evaluation_params: Default parameters to use for the validation and evaluation.
    """
    return {
                'AREA_RECALL_CONSTRAINT' : 0.8,
                'AREA_PRECISION_CONSTRAINT' : 0.4,
                'EV_PARAM_IND_CENTER_DIFF_THR': 1,
                'MTYPE_OO_O':1.,
                'MTYPE_OM_O':0.8,
                'MTYPE_OM_M':1.,
                'GT_SAMPLE_NAME_2_ID':'gt_img_([0-9]+).txt',
                'DET_SAMPLE_NAME_2_ID':'res_img_([0-9]+).txt',
                'CRLF':False # Lines are delimited by Windows CRLF format
            }

def validate_data(gtFilePath, submFilePath,evaluationParams):
    """
    Method validate_data: validates that all files in the results folder are correct (have the correct name contents).
                            Validates also that there are no missing files in the folder.
                            If some error detected, the method raises the error
    """
    gt = rrc_evaluation_funcs.load_zip_file(gtFilePath, evaluationParams['GT_SAMPLE_NAME_2_ID'])
    
    subm = rrc_evaluation_funcs.load_zip_file(submFilePath, evaluationParams['DET_SAMPLE_NAME_2_ID'], True)

    #Validate format of GroundTruth
    for k in gt:
        rrc_evaluation_funcs.validate_lines_in_file(k,gt[k],evaluationParams['CRLF'],True,True)

    #Validate format of results
    for k in subm:
        if (k in gt) == False :
            raise Exception("The sample %s not present in GT" %k)
        
        rrc_evaluation_funcs.validate_lines_in_file(k,subm[k],evaluationParams['CRLF'],True,False)

def evaluate_method(gtFilePath, submFilePath, evaluationParams):
    """
    Method evaluate_method: evaluate method and returns the results
        Results. Dictionary with the following values:
        - method (required)  Global method metrics. Ex: { 'Precision':0.8,'Recall':0.9 }
        - samples (optional) Per sample metrics. Ex: {'sample1' : { 'Precision':0.8,'Recall':0.9 } , 'sample2' : { 'Precision':0.8,'Recall':0.9 }
    """    
    
    for module,alias in evaluation_imports().items():
        globals()[alias] = importlib.import_module(module)    
    
    def one_to_one_match(row, col):
        cont = 0
        for j in range(len(recallMat[0])):    
            if recallMat[row,j] >= evaluationParams['AREA_RECALL_CONSTRAINT'] and precisionMat[row,j] >= evaluationParams['AREA_PRECISION_CONSTRAINT'] :
                cont = cont +1
        if (cont != 1):
            return False
        cont = 0
        for i in range(len(recallMat)):    
            if recallMat[i,col] >= evaluationParams['AREA_RECALL_CONSTRAINT'] and precisionMat[i,col] >= evaluationParams['AREA_PRECISION_CONSTRAINT'] :
                cont = cont +1
        if (cont != 1):
            return False
        
        if recallMat[row,col] >= evaluationParams['AREA_RECALL_CONSTRAINT'] and precisionMat[row,col] >= evaluationParams['AREA_PRECISION_CONSTRAINT'] :
            return True
        return False
    
    def one_to_many_match(gtNum):
        many_sum = 0
        detRects = []
        for detNum in range(len(recallMat[0])):    
            if gtRectMat[gtNum] == 0 and detRectMat[detNum] == 0 and detNum not in detDontCareRectsNum:
                if precisionMat[gtNum,detNum] >= evaluationParams['AREA_PRECISION_CONSTRAINT'] :
                    many_sum += recallMat[gtNum,detNum]
                    detRects.append(detNum)
        if many_sum>=evaluationParams['AREA_RECALL_CONSTRAINT'] :
            return True,detRects
        else:
            return False,[]         
    
    def many_to_one_match(detNum):
        many_sum = 0
        gtRects = []
        for gtNum in range(len(recallMat)):    
            if gtRectMat[gtNum] == 0 and detRectMat[detNum] == 0 and gtNum not in gtDontCareRectsNum:
                if recallMat[gtNum,detNum] >= evaluationParams['AREA_RECALL_CONSTRAINT'] :
                    many_sum += precisionMat[gtNum,detNum]
                    gtRects.append(gtNum)
        if many_sum>=evaluationParams['AREA_PRECISION_CONSTRAINT'] :
            return True,gtRects
        else:
            return False,[]
    
    def area(a, b):
            dx = min(a.xmax, b.xmax) - max(a.xmin, b.xmin) + 1
            dy = min(a.ymax, b.ymax) - max(a.ymin, b.ymin) + 1
            if (dx>=0) and (dy>=0):
                    return dx*dy
            else:
                    return 0.
			
    def center(r):
        x = float(r.xmin) + float(r.xmax - r.xmin + 1) / 2.;
        y = float(r.ymin) + float(r.ymax - r.ymin + 1) / 2.;
        return Point(x,y)
        
    def point_distance(r1, r2):
        distx = math.fabs(r1.x - r2.x)
        disty = math.fabs(r1.y - r2.y)
        return math.sqrt(distx * distx + disty * disty )  
        
    def center_distance(r1, r2):
        return point_distance(center(r1), center(r2))
    
    def diag(r):
        w = (r.xmax - r.xmin + 1)
        h = (r.ymax - r.ymin + 1)
        return math.sqrt(h * h + w * w)  
    
    perSampleMetrics = {}
    
    methodRecallSum = 0
    methodPrecisionSum = 0
    
    Rectangle = namedtuple('Rectangle', 'xmin ymin xmax ymax')
    Point = namedtuple('Point', 'x y')

    # 整体参数的解析
    if isinstance(gtFilePath, str):
        gt = rrc_evaluation_funcs.load_zip_file(gtFilePath,evaluationParams['GT_SAMPLE_NAME_2_ID'])
    else:
        gt = gtFilePath
    if isinstance(submFilePath, str):
        subm = rrc_evaluation_funcs.load_zip_file(submFilePath,evaluationParams['DET_SAMPLE_NAME_2_ID'],True)
    else:
        subm = submFilePath

    numGt = 0;
    numDet = 0;
   
    for resFile in gt:
        if isinstance(gt[resFile], bytes):
            gtFile = rrc_evaluation_funcs.decode_utf8(gt[resFile])
        else:
            gtFile = gt[resFile]
        recall = 0
        precision = 0
        hmean = 0        
        recallAccum = 0.
        precisionAccum = 0.
        gtRects = []
        detRects = []
        gtPolPoints = []
        detPolPoints = []
        gtDontCareRectsNum = []#Array of Ground Truth Rectangles' keys marked as don't Care
        detDontCareRectsNum = []#Array of Detected Rectangles' matched with a don't Care GT
        pairs = []
        evaluationLog = ""
        
        recallMat = np.empty([1,1])
        precisionMat = np.empty([1,1])        
            
        pointsList,_,transcriptionsList = rrc_evaluation_funcs.get_tl_line_values_from_file_contents(gtFile,evaluationParams['CRLF'],True,True,False)
        for n in range(len(pointsList)):
            points = pointsList[n]
            transcription = transcriptionsList[n]
            dontCare = transcription == "###"
            gtRect = Rectangle(*points)
            gtRects.append(gtRect)
            gtPolPoints.append(points)
            if dontCare:
                gtDontCareRectsNum.append( len(gtRects)-1 )                 
        
        evaluationLog += "GT rectangles: " + str(len(gtRects)) + (" (" + str(len(gtDontCareRectsNum)) + " don't care)\n" if len(gtDontCareRectsNum)>0 else "\n")
        
        if resFile in subm:
            if isinstance(subm[resFile], bytes):
                detFile = rrc_evaluation_funcs.decode_utf8(subm[resFile])
            else:
                detFile = subm[resFile]
            pointsList,_,_ = rrc_evaluation_funcs.get_tl_line_values_from_file_contents(detFile,evaluationParams['CRLF'],True,False,False)
            for n in range(len(pointsList)):
                points = pointsList[n]            
                detRect = Rectangle(*points)
                detRects.append(detRect)
                detPolPoints.append(points)
                if len(gtDontCareRectsNum)>0 :
                    for dontCareRectNum in gtDontCareRectsNum:
                        dontCareRect = gtRects[dontCareRectNum]
                        intersected_area = area(dontCareRect,detRect)
                        rdDimensions = ( (detRect.xmax - detRect.xmin+1) * (detRect.ymax - detRect.ymin+1));
                        if (rdDimensions==0) :
                            precision = 0
                        else:
                            precision= intersected_area / rdDimensions
                        if (precision > evaluationParams['AREA_PRECISION_CONSTRAINT'] ):
                            detDontCareRectsNum.append( len(detRects)-1 )
                            break
                             
            evaluationLog += "DET rectangles: " + str(len(detRects)) + (" (" + str(len(detDontCareRectsNum)) + " don't care)\n" if len(detDontCareRectsNum)>0 else "\n")

            if len(gtRects)==0:
                recall = 1
                precision = 0 if len(detRects)>0 else 1

            if len(detRects)>0:
                #Calculate recall and precision matrixs
                outputShape=[len(gtRects),len(detRects)]
                recallMat = np.empty(outputShape)
                precisionMat = np.empty(outputShape)
                gtRectMat = np.zeros(len(gtRects),np.int8)
                detRectMat = np.zeros(len(detRects),np.int8)
                for gtNum in range(len(gtRects)):
                    for detNum in range(len(detRects)):
                        rG = gtRects[gtNum]
                        rD = detRects[detNum]
                        intersected_area = area(rG,rD)
                        rgDimensions = ( (rG.xmax - rG.xmin+1) * (rG.ymax - rG.ymin+1) );
                        rdDimensions = ( (rD.xmax - rD.xmin+1) * (rD.ymax - rD.ymin+1));
                        recallMat[gtNum,detNum] = 0 if rgDimensions==0 else  intersected_area / rgDimensions
                        precisionMat[gtNum,detNum] = 0 if rdDimensions==0 else intersected_area / rdDimensions

                # Find one-to-one matches
                evaluationLog += "Find one-to-one matches\n"
                for gtNum in range(len(gtRects)):
                    for detNum in range(len(detRects)):
                        if gtRectMat[gtNum] == 0 and detRectMat[detNum] == 0 and gtNum not in gtDontCareRectsNum and detNum not in detDontCareRectsNum :
                            match = one_to_one_match(gtNum, detNum)
                            if match is True :
                                rG = gtRects[gtNum]
                                rD = detRects[detNum]
                                normDist = center_distance(rG, rD);
                                normDist /= diag(rG) + diag(rD);
                                normDist *= 2.0;
                                if normDist < evaluationParams['EV_PARAM_IND_CENTER_DIFF_THR'] :
                                    gtRectMat[gtNum] = 1
                                    detRectMat[detNum] = 1
                                    recallAccum += evaluationParams['MTYPE_OO_O']
                                    precisionAccum += evaluationParams['MTYPE_OO_O']
                                    pairs.append({'gt':gtNum,'det':detNum,'type':'OO'})
                                    evaluationLog += "Match GT #" + str(gtNum) + " with Det #" + str(detNum) + "\n"
                                else:
                                    evaluationLog += "Match Discarded GT #" + str(gtNum) + " with Det #" + str(detNum) + " normDist: " + str(normDist) + " \n"
                # Find one-to-many matches
                evaluationLog += "Find one-to-many matches\n"
                for gtNum in range(len(gtRects)):
                    if gtNum not in gtDontCareRectsNum:
                        match,matchesDet = one_to_many_match(gtNum)
                        if match is True :
                            gtRectMat[gtNum] = 1
                            recallAccum += evaluationParams['MTYPE_OM_O']
                            precisionAccum += evaluationParams['MTYPE_OM_O']*len(matchesDet)
                            pairs.append({'gt':gtNum,'det':matchesDet,'type':'OM'})
                            for detNum in matchesDet :
                                detRectMat[detNum] = 1
                            evaluationLog += "Match GT #" + str(gtNum) + " with Det #" + str(matchesDet) + "\n"                                

                # Find many-to-one matches
                evaluationLog += "Find many-to-one matches\n"
                for detNum in range(len(detRects)):
                    if detNum not in detDontCareRectsNum:
                        match,matchesGt = many_to_one_match(detNum)
                        if match is True :
                            detRectMat[detNum] = 1
                            recallAccum += evaluationParams['MTYPE_OM_M']*len(matchesGt)
                            precisionAccum += evaluationParams['MTYPE_OM_M']
                            pairs.append({'gt':matchesGt,'det':detNum,'type':'MO'})
                            for gtNum in matchesGt :
                                gtRectMat[gtNum] = 1
                            evaluationLog += "Match GT #" + str(matchesGt) + " with Det #" + str(detNum) + "\n"

                numGtCare = (len(gtRects) - len(gtDontCareRectsNum))
                if numGtCare == 0:
                    recall = float(1)
                    precision = float(0) if len(detRects)>0 else float(1)
                else:
                    recall = float(recallAccum) / numGtCare
                    precision =  float(0) if (len(detRects) - len(detDontCareRectsNum))==0 else float(precisionAccum) / (len(detRects) - len(detDontCareRectsNum))
                hmean = 0 if (precision + recall)==0 else 2.0 * precision * recall / (precision + recall)  
                
        evaluationLog += "Recall = " + str(recall) + "\n"
        evaluationLog += "Precision = " + str(precision) + "\n"

        methodRecallSum += recallAccum
        methodPrecisionSum += precisionAccum
        numGt += len(gtRects) - len(gtDontCareRectsNum)
        numDet += len(detRects) - len(detDontCareRectsNum)

        perSampleMetrics[resFile] = {
                                        'precision':precision,
                                        'recall':recall,
                                        'hmean':hmean,
                                        'pairs':pairs,
                                        'recallMat': [] if len(detRects)>100 else recallMat.tolist(),
                                        'precisionMat':[] if len(detRects)>100 else precisionMat.tolist(),
                                        'gtPolPoints':gtPolPoints,
                                        'detPolPoints':detPolPoints,
                                        'gtDontCare':gtDontCareRectsNum,
                                        'detDontCare':detDontCareRectsNum,
                                        'evaluationParams': evaluationParams,
                                        'evaluationLog': evaluationLog
                                    }
        
    methodRecall = 0 if numGt==0 else methodRecallSum/numGt
    methodPrecision = 0 if numDet==0 else methodPrecisionSum/numDet
    methodHmean = 0 if methodRecall + methodPrecision==0 else 2* methodRecall * methodPrecision / (methodRecall + methodPrecision)
    
    methodMetrics = {'precision':methodPrecision, 'recall':methodRecall,'hmean': methodHmean  }

    resDict = {'calculated':True,'Message':'','method': methodMetrics,'per_sample': perSampleMetrics}
    
    
    return resDict;



if __name__=='__main__':
        
    rrc_evaluation_funcs.main_evaluation(None,default_evaluation_params,validate_data,evaluate_method)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/03/03 09:35

""" 主要含有功能

结果分数测评
结果调试、可视化分析
"""

import pandas as pd

from pyxlpr.data.coco import *


class KieResParser:
    """ Key Information Extraction Result Parser

    关键信息提取结果的分析工具
    """

    def __init__(self, df, keys):
        """
        :param df: 必须要有一列file文件标识，用来进行分组判断，每一组的识别效果
        :param keys: 信息种类，有哪些键
            for k in keys，都有对应名称的 dt_{k},  gt_{k}  属性列，测试集场合，可以没有 gt 列
        """
        assert 'file' in df.columns
        self.df = df
        self.keys = keys
        self.add_diff_columns()

    def add_diff_columns(self):
        """ 使用该功能，必须要有对应的gt列 """
        for k in self.keys:
            diff_col = f'diff_{k}'
            if diff_col not in self.df.columns:
                dt_col, gt_col = f'dt_{k}', f'gt_{k}'
                self.df[diff_col] = [StrDiffType.difftype(row[dt_col], row[gt_col]) for idx, row in self.df.iterrows()]

    def check_file(self):
        """ 按每份文档为单位查看效果

        TODO
            1、支持查看单文件
            2、支持筛选key，只对比查看部分key
            3、支持显示key名称
        """
        pc = PairContent(Dir.TEMP / 'dt.txt', Dir.TEMP / 'gt.txt')
        for i, (idx, item) in enumerate(self.df.iterrows(), start=1):
            pc.add(f'{i}、{item.file}')
            for k in self.keys:
                pc.add(item[f'dt_{k}'], item[f'gt_{k}'])
            pc.add('')
        pc.bcompare(wait=False)

    def stat_difftype(self):
        """ 分析dt各种情况下检测效果、质量 """
        items = []
        index = []
        for i in StrDiffType.typename.keys():
            vals = []
            for k in self.keys:
                vals.append(sum(self.df[f'diff_{k}'] == i))
            if sum(vals):
                items.append(vals)
                index.append(i)
        df = pd.DataFrame.from_records(items, columns=self.keys, index=index)
        return df

    def check_key(self, key):
        """ 检查某个特定key的识别效果 """
        n = len(self.df)  # 总条目数
        pc = PairContent(Dir.TEMP / f'dt-{key}.txt', Dir.TEMP / f'gt-{key}.txt')
        dt_col, gt_col, diff_col = f'dt_{key}', f'gt_{key}', f'diff_{key}'
        for difftype, items in self.df.groupby(diff_col):
            m = len(items)
            pc.add(f'{difftype}、{StrDiffType.typename[difftype]}（{m}/{n}≈{m / n:.2%}）')
            for _, x in items.iterrows():
                f = x['file'] + ', '
                pc.add(f + x[dt_col], f + x[gt_col])
            pc.add('')
        pc.bcompare(wait=False)



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This code is refer from:
https://github.com/RubanSeven/Text-Image-Augmentation-python/blob/master/warp_mls.py
"""

import numpy as np


class WarpMLS:
    def __init__(self, src, src_pts, dst_pts, dst_w, dst_h, trans_ratio=1.):
        self.src = src
        self.src_pts = src_pts
        self.dst_pts = dst_pts
        self.pt_count = len(self.dst_pts)
        self.dst_w = dst_w
        self.dst_h = dst_h
        self.trans_ratio = trans_ratio
        self.grid_size = 100
        self.rdx = np.zeros((self.dst_h, self.dst_w))
        self.rdy = np.zeros((self.dst_h, self.dst_w))

    @staticmethod
    def __bilinear_interp(x, y, v11, v12, v21, v22):
        return (v11 * (1 - y) + v12 * y) * (1 - x) + (v21 *
                                                      (1 - y) + v22 * y) * x

    def generate(self):
        self.calc_delta()
        return self.gen_img()

    def calc_delta(self):
        w = np.zeros(self.pt_count, dtype=np.float32)

        if self.pt_count < 2:
            return

        i = 0
        while 1:
            if self.dst_w <= i < self.dst_w + self.grid_size - 1:
                i = self.dst_w - 1
            elif i >= self.dst_w:
                break

            j = 0
            while 1:
                if self.dst_h <= j < self.dst_h + self.grid_size - 1:
                    j = self.dst_h - 1
                elif j >= self.dst_h:
                    break

                sw = 0
                swp = np.zeros(2, dtype=np.float32)
                swq = np.zeros(2, dtype=np.float32)
                new_pt = np.zeros(2, dtype=np.float32)
                cur_pt = np.array([i, j], dtype=np.float32)

                k = 0
                for k in range(self.pt_count):
                    if i == self.dst_pts[k][0] and j == self.dst_pts[k][1]:
                        break

                    w[k] = 1. / (
                        (i - self.dst_pts[k][0]) * (i - self.dst_pts[k][0]) +
                        (j - self.dst_pts[k][1]) * (j - self.dst_pts[k][1]))

                    sw += w[k]
                    swp = swp + w[k] * np.array(self.dst_pts[k])
                    swq = swq + w[k] * np.array(self.src_pts[k])

                if k == self.pt_count - 1:
                    pstar = 1 / sw * swp
                    qstar = 1 / sw * swq

                    miu_s = 0
                    for k in range(self.pt_count):
                        if i == self.dst_pts[k][0] and j == self.dst_pts[k][1]:
                            continue
                        pt_i = self.dst_pts[k] - pstar
                        miu_s += w[k] * np.sum(pt_i * pt_i)

                    cur_pt -= pstar
                    cur_pt_j = np.array([-cur_pt[1], cur_pt[0]])

                    for k in range(self.pt_count):
                        if i == self.dst_pts[k][0] and j == self.dst_pts[k][1]:
                            continue

                        pt_i = self.dst_pts[k] - pstar
                        pt_j = np.array([-pt_i[1], pt_i[0]])

                        tmp_pt = np.zeros(2, dtype=np.float32)
                        tmp_pt[0] = np.sum(pt_i * cur_pt) * self.src_pts[k][0] - \
                                    np.sum(pt_j * cur_pt) * self.src_pts[k][1]
                        tmp_pt[1] = -np.sum(pt_i * cur_pt_j) * self.src_pts[k][0] + \
                                    np.sum(pt_j * cur_pt_j) * self.src_pts[k][1]
                        tmp_pt *= (w[k] / miu_s)
                        new_pt += tmp_pt

                    new_pt += qstar
                else:
                    new_pt = self.src_pts[k]

                self.rdx[j, i] = new_pt[0] - i
                self.rdy[j, i] = new_pt[1] - j

                j += self.grid_size
            i += self.grid_size

    def gen_img(self):
        src_h, src_w = self.src.shape[:2]
        dst = np.zeros_like(self.src, dtype=np.float32)

        for i in np.arange(0, self.dst_h, self.grid_size):
            for j in np.arange(0, self.dst_w, self.grid_size):
                ni = i + self.grid_size
                nj = j + self.grid_size
                w = h = self.grid_size
                if ni >= self.dst_h:
                    ni = self.dst_h - 1
                    h = ni - i + 1
                if nj >= self.dst_w:
                    nj = self.dst_w - 1
                    w = nj - j + 1

                di = np.reshape(np.arange(h), (-1, 1))
                dj = np.reshape(np.arange(w), (1, -1))
                delta_x = self.__bilinear_interp(
                    di / h, dj / w, self.rdx[i, j], self.rdx[i, nj],
                    self.rdx[ni, j], self.rdx[ni, nj])
                delta_y = self.__bilinear_interp(
                    di / h, dj / w, self.rdy[i, j], self.rdy[i, nj],
                    self.rdy[ni, j], self.rdy[ni, nj])
                nx = j + dj + delta_x * self.trans_ratio
                ny = i + di + delta_y * self.trans_ratio
                nx = np.clip(nx, 0, src_w - 1)
                ny = np.clip(ny, 0, src_h - 1)
                nxi = np.array(np.floor(nx), dtype=np.int32)
                nyi = np.array(np.floor(ny), dtype=np.int32)
                nxi1 = np.array(np.ceil(nx), dtype=np.int32)
                nyi1 = np.array(np.ceil(ny), dtype=np.int32)

                if len(self.src.shape) == 3:
                    x = np.tile(np.expand_dims(ny - nyi, axis=-1), (1, 1, 3))
                    y = np.tile(np.expand_dims(nx - nxi, axis=-1), (1, 1, 3))
                else:
                    x = ny - nyi
                    y = nx - nxi
                dst[i:i + h, j:j + w] = self.__bilinear_interp(
                    x, y, self.src[nyi, nxi], self.src[nyi, nxi1],
                    self.src[nyi1, nxi], self.src[nyi1, nxi1])

        dst = np.clip(dst, 0, 255)
        dst = np.array(dst, dtype=np.uint8)

        return dst


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2020/05/30 21:14

"""
百度人工智能API，常用URL
使用文档：https://cloud.baidu.com/doc/OCR/s/Ek3h7xypm
调用次数：https://console.bce.baidu.com/ai/?_=1653139065257#/ai/ocr/overview/index


"""
import os
import base64
import json
import pprint
import statistics
import time

import cv2
import numpy as np
import requests

from pyxllib.prog.newbie import round_int
from pyxllib.prog.pupil import check_install_package, is_url
from pyxllib.prog.xlosenv import XlOsEnv
from pyxllib.algo.geo import xywh2ltrb, rect_bounds
from pyxllib.file.specialist import XlPath, get_etag
from pyxllib.cv.expert import xlcv


def __1_转类labelme标注():
    """ 将百度api获得的各种业务结果，转成类似labelme的标注格式

    统一成一种风格，易分析的结构
    """


def loc2points(loc, ratio=1):
    """ 百度的location格式转为 {'points': [[a, b], [c, d]], 'shape_type: 'rectangle'} """
    # 目前关注到的都是矩形，不知道有没可能有其他格式
    ltrb = xywh2ltrb([loc['left'], loc['top'], loc['width'], loc['height']])
    if ratio != 1:
        ltrb = [x * ratio for x in ltrb]
    l, t, r, b = round_int(ltrb, ndim=1)
    return {'points': [[l, t], [r, b]],
            'shape_type': 'rectangle'}


def loc2points2(loc, ratio=1):
    """ 百度的location格式转为 {'points': [[a, b], [c, d]], 'shape_type: 'rectangle'} """
    # 目前关注到的都是矩形，不知道有没可能有其他格式
    ltrb = xywh2ltrb([loc['left'], loc['top'], loc['right'] - loc['left'], loc['bottom'] - loc['top']])
    if ratio != 1:
        ltrb = [x * ratio for x in ltrb]
    l, t, r, b = round_int(ltrb, ndim=1)
    return {'points': [[l, t], [r, b]],
            'shape_type': 'rectangle'}


def polygon2rect(pts):
    l, t, r, b = rect_bounds(pts)
    return [[l, t], [r, b]]


def zoom_point(pt, ratio):
    """ 返回的还是{x: 123, y: 456}的字典结构 """
    if ratio != 1:
        return {k: round_int(v * ratio) for k, v in pt.items()}
    return pt


def zoom_point2(pt, ratio):
    """ 输入x,y的字典坐标，返回的还是[123, 456]的list结构 """
    pt = [pt['x'], pt['y']]
    if ratio != 1:
        return [round_int(v * ratio) for v in pt]
    return pt


def zoom_labelme(d, ratio):
    """ 对labelme的标注进行缩放 """
    ratio2 = 1 / ratio
    for sp in d['shapes']:
        sp['points'] = [[round_int(p[0] * ratio2), round_int(p[1] * ratio2)] for p in sp['points']]
    d['imageHeight'] = round_int(d['imageHeight'] * ratio2)
    d['imageWidth'] = round_int(d['imageWidth'] * ratio2)
    return d


def labelmelike_extend_args(core_func):
    """ 扩展 main_func 函数，支持一些通用上下游切面功能

    支持 main_keys，主要数据字段名称
    支持 remove_keys，批量移除一些不需要的参数值
    """

    def wrapper(data, ratio, main_key, remove_keys=None, *, clear_empty_shape=False):
        """
        :param clear_empty_shape: 在考虑要不要把清除空shape，四边形转矩形等功能写到这里，暂时还未写对应功能
        """
        # 1 转主要的数据结构
        if main_key in data:
            data['shapes'] = core_func(data[main_key], ratio)

        # 2 删除不需要的键值
        if remove_keys is None:
            _remove_keys = set()
        elif not isinstance(remove_keys, (list, tuple, set)):
            _remove_keys = [remove_keys]
        else:
            _remove_keys = remove_keys
        for k in ({main_key} | set(_remove_keys)):
            if k in data:
                del data[k]

        return data

    return wrapper


class ToLabelmeLike:
    """ 注意配合装饰器的使用
    这里每个函数，只要实现输入核心识别结果清单，返回shapes的代码即可。上下游一些切面操作是统一的。
    """

    @staticmethod
    @labelmelike_extend_args
    def list_word(ls, ratio):
        shapes = []
        for w in ls:
            shape = {}
            if 'location' in w:
                shape.update(loc2points(w['location'], ratio))
            if 'words' in w:
                # shape['label'] = json.dumps({'text': x['words']}, ensure_ascii=False)
                shape['label'] = {'text': w['words']}  # 不转字符串
            shapes.append(shape)
        return shapes

    @staticmethod
    @labelmelike_extend_args
    def list_word2(ls, ratio):
        shapes = []
        for w in ls:
            shape = {}
            shape.update(loc2points(w['words']['words_location'], ratio))
            shape['label'] = {'text': w['words']['word'], 'category': w['words_type']}
            shapes.append(shape)
        return shapes

    @staticmethod
    @labelmelike_extend_args
    def dict_word(d, ratio):
        shapes = []
        for k, w in d.items():
            shape = {'label': {'category': k}}
            if 'location' in w:
                shape.update(loc2points(w['location'], ratio))
            if 'words' in w:
                shape['label']['text'] = w['words']
            shapes.append(shape)
        return shapes

    @staticmethod
    @labelmelike_extend_args
    def dict_str(d, ratio):
        return [{'label': {'text': v, 'category': k}} for k, v in d.items()]

    @staticmethod
    @labelmelike_extend_args
    def dict_strs(d, ratio):
        shapes = []
        for k, texts in d.items():
            shapes.append({'label': {'category': k, 'text': ','.join(texts)}})
        return shapes


def __2_定制不同输出格式():
    pass


class XlAiClient:
    """
    封装该类
        目的1：合并输入文件和url的识别
        目的2：带透明底的png百度api识别不了，要先转成RGB格式
    """
    META_OPTS_NAME = {'record_files',  # 是否记录传入的图片等数据
                      'record_xlapi',  # 是否记录api处理结果
                      'use_exists_xlapi',  # 是否使用已有的api结果
                      'record_xlserver',  # 是否记录调用纪录
                      }

    def __init__(self, auto_login=True, check=True):
        # 1 默认值
        self.db = None

        self._aipocr = None

        self._mathpix_header = None

        self._priu_header = None
        self._priu_host = None

        # 2 如果环境变量预存了账号信息，自动加载使用
        accounts = XlOsEnv.get('XlAiAccounts', decoding=True)
        if accounts and auto_login:
            if 'aipocr' in accounts:
                self.login_aipocr(**accounts['aipocr'])
            if 'mathpix' in accounts:
                self.login_mathpix(**accounts['mathpix'])
            if 'priu' in accounts:
                self.login_priu(**accounts['priu'], check=check)

    def __A1_登录账号(self):
        pass

    def set_database(self, db):
        """ 是否关联数据库，查找已运行过的结果，或存储运行结果

        from pyxllib.data.pglib import XlprDb
        db = XlprDb.connect()
        self.setup_database(db)
        """
        self.db = db

    def login_aipocr(self, app_id, api_key, secret_key):
        """
        注：带透明底的png百度api识别不了，要先转成RGB格式
        """
        check_install_package('aip', 'baidu-aip')
        import aip

        self._aipocr = aip.AipOcr(str(app_id), api_key, secret_key)

    def login_mathpix(self, app_id, app_key):
        self._mathpix_header = {'Content-type': 'application/json'}
        self._mathpix_header.update({'app_id': app_id, 'app_key': app_key})

    def login_priu(self, token, host=None, *, check=True):
        """ 福建省模式识别与图像理解重点实验室

        :param str|list[str] host:
            str, 主机IP，比如'118.195.202.82'
            list, 也可以输入一个列表，会从左到右依次尝试链接，使用第一个能成功链接的ip，常用语优先选用局域网，找不到再使用公网接口
        """
        self._priu_header = {'Content-type': 'application/json'}
        self._priu_header.update({'Token': token})

        if host is None:
            # 优先尝试局域网链接，如果失败则尝试公网链接。公网默认用https，如果需要效率可以自己输入http的host。
            hosts = ['172.16.170.136', 'https://xmutpriu.com']
        elif isinstance(host, str):
            hosts = [host]
        elif isinstance(host, (list, tuple)):
            hosts = host
        else:
            raise TypeError

        # 确保都有http或https前缀
        for i, host in enumerate(hosts):
            if 'http' not in host:
                hosts[i] = f'http://{host}'

        if check:
            connent = False
            for host in hosts:
                try:
                    if '欢迎来到 厦门理工' in requests.get(f'{host}/test_page', timeout=5).text:
                        self._priu_host = host
                        connent = True
                        break
                except (requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout):
                    continue

            if not connent:
                raise requests.exceptions.ConnectionError('PRIU接口登录失败')
            return connent
        else:
            self._priu_host = hosts[0]  # 不检查的场景，一般都是局域网内使用
            return None

    def __A2_调整图片和关联数据库(self):
        pass

    @classmethod
    def adjust_image(cls, in_, flags=1, *, b64decode=True, to_buffer=True, b64encode=False,
                     min_length=15, max_length=4096,
                     limit_b64buffer_size=4 * 1024 ** 2):
        """ 这里是使用API接口，比较通用的一套图片处理操作

        :param in_: 可以是本地文件，也可以是图片url地址，也可以是Image对象
            注意这个函数，输入是url，也会获取重置图片数据上传
            如果为了效率明确只传url，可以用aip.AipOcr原生的相关url函数
        :param b64decode: 如果输入是bytes类型，是否要用b64解码，默认需要
        :return: 返回图片文件二进制值的buffer, 缩放系数(小余1是缩小，大于1是放大)
        """
        # 1 取不同来源的数据
        # 下面应该是比较通用的一套操作，如果有特殊接口，可以另外处理，不一定要通过该接口处理图片
        if isinstance(in_, bytes):
            im = xlcv.read_from_buffer(in_, flags, b64decode=b64decode)
        elif is_url(in_):
            im = xlcv.read_from_url(in_, flags)
        else:
            im = xlcv.read(in_, flags)
        origin_height = im.shape[0]

        if im.dtype == 'uint16':
            im = cv2.convertScaleAbs(im)

        # 2 图片尺寸不符合要求，要缩放
        if min_length or max_length:
            im = xlcv.adjust_shape(im, min_length, max_length)

        # 3 图片文件不能过大，要调整
        if limit_b64buffer_size:
            # b64后大小固定会变4/3，所以留给原文件的大小是要缩水，只有0.75；再以防万一总不能卡得刚刚好，所以设为0.74
            im = xlcv.reduce_filesize(im, limit_b64buffer_size * 0.74)
        current_height = im.shape[0]

        if to_buffer:
            im = xlcv.to_buffer(im, '.jpg', b64encode=b64encode)
        ratio = current_height / origin_height
        return im, ratio

    def _divide_options_with_meta(self, options, meta_opts):
        """ 会改变原始输入的两个字典值，但是新返回的才是真正有效的字典 """
        for name in self.META_OPTS_NAME:
            if name in options:
                meta_opts[name] = options[name]
                del options[name]

        options = options or {}
        options = {k: options[k] for k in sorted(options.keys())}  # 对参数进行排序，方便记录到数据库后标记唯一

        return options, meta_opts

    def run_aipocr_with_db(self, func, buffer, options=None, **meta_opts):
        """ 这一套主要是适配百度视觉api接口的情况，基本都是输入一个buffer和一个options
        如果自己模型有特殊接口范式，可以额外定制，比如 run_textapi_with_db 等，可以拷贝这个函数后做定制微调

        :param func: 被封装的带执行的api函数
        :param buffer: 图片二进制数据
        :param options: 执行api功能的配套采纳数
        :param meta_opts:
            record_files: 是否保存数据文件
                False, 0，表示不存储
                True，表示存储，但默认只存储4MB以内的文件
                int，可以写一个整数，表示只存储多大Byte范围内的文件
            record_xlapi: 新结果是否记录到数据库
            use_exists_xlapi: 如果数据库里已有记录，是否直接复用
                注意看函数实现，但这个机制生效已经获得res的时候，变不会显式执行record_files
                这是出于效率的考虑，一般情况下，这种应该是已经存储过文件了
                如果要怕以前执行过api但没存储files，可以关闭use_exists_xlapi，就可以显式重新执行保存一次
            mode_name: 可以指定存入数据库的功能名，默认用func的名称
        """
        # 1 参数处理
        options, meta_opts = self._divide_options_with_meta(options, meta_opts)
        if self.db is None:  # 没有数据库的时候直接执行函数返回结果就好
            return func(buffer, options)

        # 以下元参数默认值，根据不同的api接口需求，可以设置不同的默认值
        record_files = meta_opts.get('record_files', True)  # 是否保存数据文件
        if record_files is True:
            record_files = 4 * 1024 ** 2
        record_xlapi = meta_opts.get('record_xlapi', True)
        use_exists_xlapi = meta_opts.get('use_exists_xlapi', True)
        mode_name = meta_opts.get('mode_name', func.__name__)

        image_etag = None  # 先不计算etag，只在需要的时候再计算
        res = None  # api结果

        # 2 核心函数
        if use_exists_xlapi:  # 定义如何获取已有的识别结果
            image_etag = image_etag or get_etag(buffer)  # trick: 使用这个机制确保etag从头到尾不会重复计算，需要的话只会计算一次
            res = self.db.get_xlapi_record(mode=mode_name, image=image_etag, **options)

        # 否则调用百度的接口识别
        # TODO 这里使用协程逻辑最合理但配置麻烦，需要func底层等做协程的适配支持
        #   使用多线程测试了并没有更快，也发现主要耗时是post，数据库不会花太多时间，就先不改动了
        #   等以后数据库大了，看运行是否会慢，可以再测试是否有必要弄协程
        if res is None or 'error_code' in res:
            tt = time.time()
            res = func(buffer, options)
            elapse_ms = round_int(1000 * (time.time() - tt))

            if len(buffer) < record_files:  # 定义数据的缓存方式
                image_etag = image_etag or get_etag(buffer)
                self.db.insert_row2files(buffer, etag=image_etag, name='.jpg')

            if record_xlapi:  # 定义api的执行结果的保存方式
                image_etag = image_etag or get_etag(buffer)
                input = {'mode': mode_name, 'image': image_etag}
                if options:
                    input.update(options)
                xlapi_id = self.db.insert_row2xlapi(input, res, elapse_ms, on_conflict='REPLACE')
                res['xlapi_id'] = xlapi_id

        # 3 收尾
        if 'log_id' in res:  # 有xlapi_id的标记，就不用百度原本的log_id了
            del res['log_id']

        return res

    def __B1_通用(self):
        pass

    def general(self, image, **options):
        """ 通用文字识别（标准含位置版）: https://cloud.baidu.com/doc/OCR/s/vk3h7y58v
        500次/天赠送 + 超出按量计费

        注意只要ratio!=1涉及到缩放的，偏移误差是会变大的~~
        """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.general, buffer, options)
        result_dict = ToLabelmeLike.list_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def basicGeneral(self, image, **options):
        """ 通用文字识别（标准版）: https://cloud.baidu.com/doc/OCR/s/zk3h7xz52
        5万次/天赠送 + 超出按量计费
        """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.basicGeneral, buffer, options)
        result_dict = ToLabelmeLike.list_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def accurate(self, image, **options):
        """ 通用文字识别（高精度含位置版）: https://cloud.baidu.com/doc/OCR/s/tk3h7y2aq """
        sz = 10 * 1024 ** 2
        buffer, ratio = self.adjust_image(image, max_length=8192, limit_b64buffer_size=sz, to_buffer=True)
        result_dict = self.run_aipocr_with_db(self._aipocr.accurate, buffer, options, save_buffer_threshold_size=sz)
        result_dict = ToLabelmeLike.list_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def basicAccurate(self, image, **options):
        """ 通用文字识别（高精度版）: https://cloud.baidu.com/doc/OCR/s/1k3h7y3db """
        sz = 10 * 1024 ** 2
        buffer, ratio = self.adjust_image(image, max_length=8192, limit_b64buffer_size=sz, to_buffer=True)
        result_dict = self.run_aipocr_with_db(self._aipocr.basicAccurate, buffer, options, save_buffer_threshold_size=sz)
        result_dict = ToLabelmeLike.list_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def webimageLoc(self, image, **options):
        """ 网络图片文字识别（含位置版）: https://cloud.baidu.com/doc/OCR/s/Nkaz574we """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.webimageLoc, buffer, options)
        result_dict = ToLabelmeLike.list_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def webImage(self, image, **options):
        """ 网络图片文字识别: https://cloud.baidu.com/doc/OCR/s/Sk3h7xyad """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.webImage, buffer, options)
        result_dict = ToLabelmeLike.list_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def handwriting(self, image, **options):
        """ 手写文字识别: https://cloud.baidu.com/doc/OCR/s/hk3h7y2qq """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.handwriting, buffer, options)
        result_dict = ToLabelmeLike.list_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def numbers(self, image, **options):
        """ 数字识别: https://cloud.baidu.com/doc/OCR/s/Ok3h7y1vo """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.numbers, buffer, options)
        result_dict = ToLabelmeLike.list_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def qrcode(self, image, **options):
        """ 二维码识别: https://cloud.baidu.com/doc/OCR/s/qk3h7y5o7 """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.qrcode, buffer, options)

        ls = result_dict['codes_result']
        shapes = []
        for x in ls:
            # assert len(x['text']) == 1, '取到的文本list长度大于1'
            shapes.append({'label': {'text': ','.join(x['text']), 'category': x['type']}})
        result_dict['shapes'] = shapes

        del result_dict['codes_result']
        del result_dict['codes_result_num']
        return result_dict

    def form(self, image, **options):
        """ 表格文字识别(同步接口): https://cloud.baidu.com/doc/OCR/s/ik3h7xyxf

        备注：
            0、有异步接口
            1、返回是四边形结构，实际值都是矩形，先转成矩形处理了
            2、因为矩形，这个不能处理倾斜的表格
            3、看起来支持多表格
        """
        from bisect import bisect_left
        from pyxllib.file.xlsxlib import XlWorkbook

        # 1 单表处理功能
        def zoom_coords(table, ratio):
            for k, xs in table.items():
                if k == 'vertexes_location':
                    table[k] = [zoom_point(pt, ratio) for pt in table[k]]
                else:
                    for x in table[k]:
                        x['vertexes_location'] = [zoom_point(pt, ratio) for pt in x['vertexes_location']]

        def parse_split_points(cells, cls='row', name='y'):
            # 计算行列数量，初始化空list
            n = max([x[cls] for x in cells]) + 2
            spans = [[] for i in range(n)]
            # 加入所有行列数据
            for x in cells:
                spans[x[cls]].append(x['vertexes_location'][0][name])
            spans[-1].append(table['vertexes_location'][2][name])
            # 计算平均行列位置
            spans = [statistics.mean(vs) for vs in spans if vs]
            # 合并临近的行列
            min_gap = 3  # 两行/列间距小余min_gap像素，合并为1行/列
            spans = [v for a, v in zip([0] + spans, spans) if v - a > min_gap]
            # 计算分割线（以中间为准）
            spans = [round_int((a + b) / 2) for a, b in zip(spans, spans[1:])]
            return spans

        def location2rowcol(loc, rowspan, colspan):
            p1, p2 = loc[0], loc[2]
            pos = {
                'row': bisect_left(rowspan, p1['y']) + 1,  # 我的数据改成从1开始编号了
                'column': bisect_left(colspan, p1['x']) + 1,
                'end_row': bisect_left(rowspan, p2['y']),
                'end_column': bisect_left(colspan, p2['x']),
            }
            return pos

        def parse_table(table):
            # 1 计算行、列分界线
            rowspan = parse_split_points(table['body'], 'row', 'y')
            colspan = parse_split_points(table['body'], 'column', 'x')

            # 2 shapes
            shapes = []
            # 这个表格暂时转labelme格式，但泛用来说，其实不应该转labelme，它有其特殊性
            # location虽然给的是四边形，但看目前数据其实就是矩形
            for k, xs in table.items():
                if k == 'vertexes_location':
                    sp = {'label': {'text': '', 'category': 'table'},
                          'points': polygon2rect([(p['x'], p['y']) for p in table['vertexes_location']]),
                          'shape_type': 'rectangle'}
                    shapes.append(sp)
                else:
                    for x in xs:
                        label = {'text': x['words'], 'probability': x['probability'], 'category': k}
                        label.update(location2rowcol(x['vertexes_location'], rowspan, colspan))
                        sp = {'label': label,
                              'points': polygon2rect([(p['x'], p['y']) for p in x['vertexes_location']]),
                              'shape_type': 'rectangle'}
                        shapes.append(sp)

            # 3 tables
            # 3.1 表格主体内容
            wb = XlWorkbook()
            ws = wb.active
            for sp in shapes:
                x = sp['label']
                if x['category'] != 'body':
                    continue
                # ws.cell(x['row'], x['column'], x['text'])
                cel = ws.cell(x['row'], x['column']).mcell()
                cel.value = (cel.value or '') + x['text']
                if x['end_row'] - x['row'] + x['end_column'] - x['column'] > 0:
                    ws.merge_cells(start_row=x['row'], start_column=x['column'],
                                   end_row=x['end_row'], end_column=x['end_column'])
            # wb.save('/home/chenkunze/data/aipocr_test/a.xlsx')  # debug: 保存中间结果查看

            # 3.2 其他内容整体性拼接
            htmltable = ['<div>']
            header = '<br/>'.join([x['words'] for x in table['header']])
            footer = '<br/>'.join([x['words'] for x in table['footer']])
            if header:
                htmltable.append(f'<p>{header}</p>')
            htmltable.append(ws.to_html())
            if footer:
                htmltable.append(f'<p>{footer}</p>')
            htmltable.append('</div>')

            return shapes, '\n'.join(htmltable)

        # 2 主体解析功能

        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.form, buffer, options)
        shapes = []
        htmltables = []
        for table in result_dict['forms_result']:
            if ratio != 1:
                zoom_coords(table, 1 / ratio)
            _shapes, htmltable = parse_table(table)
            shapes += _shapes
            htmltables.append(htmltable)

        # 3 收尾
        result_dict['shapes'] = shapes
        result_dict['htmltables'] = htmltables
        del result_dict['forms_result']
        del result_dict['forms_result_num']
        return result_dict

    def doc_analysis_office(self, image, **options):
        """ 办公文档识别: https://cloud.baidu.com/doc/OCR/s/ykg9c09ji """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.doc_analysis_office, buffer, options)
        result_dict = ToLabelmeLike.list_word2(result_dict, 1 / ratio, 'results', 'results_num')
        return result_dict

    def seal(self, image, **options):
        """ 印章识别: https://cloud.baidu.com/doc/OCR/s/Mk3h7y47a """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.seal, buffer, options)

        shapes = []
        for x in result_dict['result']:
            shape = {'label': {}}
            shape.update(loc2points(x['location'], 1 / ratio))
            shape['label']['text'] = x['major']['words']
            shape['label']['minor'] = ','.join(x['minor'])
            shape['label']['category'] = x['type']
            shapes.append(shape)

        result_dict['shapes'] = shapes
        del result_dict['result']
        del result_dict['result_num']
        return result_dict

    def __B2_卡证(self):
        pass

    def idcard(self, image, **options):
        """ 身份证识别: https://cloud.baidu.com/doc/OCR/s/rk3h7xzck """

        def idcard_front(image, options=None):
            return self._aipocr.idcard(image,
                                       options.get('id_card_side', 'front'),  # 默认识别带照片一面
                                       options)

        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(idcard_front, buffer, options)
        result_dict = ToLabelmeLike.dict_word(result_dict, 1 / ratio, 'words_result', 'words_result')
        return result_dict

    def idcard_back(self, image, **options):
        def func(image, options=None):
            return self._aipocr.idcard(image,
                                       options.get('id_card_side', 'back'),  # 识别国徽一面
                                       options)

        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(func, buffer, options, mode_name='idcard_back')
        result_dict = ToLabelmeLike.dict_word(result_dict, 1 / ratio, 'words_result', 'words_result')
        return result_dict

    def bankcard(self, image, **options):
        """ 银行卡识别: https://cloud.baidu.com/doc/OCR/s/ak3h7xxg3 """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.bankcard, buffer, options)
        result_dict = ToLabelmeLike.dict_str(result_dict, 1 / ratio, 'result', 'words_result_num')
        return result_dict

    def businessLicense(self, image, **options):
        """ 营业执照识别: https://cloud.baidu.com/doc/OCR/s/sk3h7y3zs """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.businessLicense, buffer, options)
        result_dict = ToLabelmeLike.dict_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def businessCard(self, image, **options):
        """ 名片识别: https://cloud.baidu.com/doc/OCR/s/5k3h7xyi2 """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.businessCard, buffer, options)
        result_dict = ToLabelmeLike.dict_strs(result_dict, 1 / ratio, 'words_result')
        return result_dict

    def passport(self, image, **options):
        """ 护照识别: https://cloud.baidu.com/doc/OCR/s/Wk3h7y1gi """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.passport, buffer, options)
        result_dict = ToLabelmeLike.dict_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def HKMacauExitentrypermit(self, image, **options):
        """ 港澳通行证识别: https://cloud.baidu.com/doc/OCR/s/4k3h7y0ly """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.HKMacauExitentrypermit, buffer, options)
        result_dict = ToLabelmeLike.dict_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def taiwanExitentrypermit(self, image, **options):
        """ 台湾通行证识别: https://cloud.baidu.com/doc/OCR/s/kk3h7y2yc """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.taiwanExitentrypermit, buffer, options)
        result_dict = ToLabelmeLike.dict_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def householdRegister(self, image, **options):
        """ 户口本识别: https://cloud.baidu.com/doc/OCR/s/ak3h7xzk7 """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.householdRegister, buffer, options)
        result_dict = ToLabelmeLike.dict_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def birthCertificate(self, image, **options):
        """ 出生医学证明识别: https://cloud.baidu.com/doc/OCR/s/mk3h7y1o6 """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.birthCertificate, buffer, options)
        result_dict = ToLabelmeLike.dict_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def __B3_交通(self):
        pass

    def vehicleLicense(self, image, **options):
        """ 行驶证识别: https://cloud.baidu.com/doc/OCR/s/yk3h7y3ks """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.vehicleLicense, buffer, options)
        result_dict = ToLabelmeLike.dict_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def drivingLicense(self, image, **options):
        """ 驾驶证识别: https://cloud.baidu.com/doc/OCR/s/Vk3h7xzz7 """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.drivingLicense, buffer, options)
        result_dict = ToLabelmeLike.dict_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def licensePlate(self, image, **options):
        """ 车牌识别: https://cloud.baidu.com/doc/OCR/s/ck3h7y191 """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.licensePlate, buffer, options)

        if 'words_result' in result_dict:
            d = result_dict['words_result']
        else:
            d = {'number': '', 'color': '', 'probability': [0], 'vertexes_location': []}

        # 近似key_text，但有点不太一样
        # 已测试过 车牌 只能识别一个
        label = {'text': d['number'],
                 'color': d['color']}
        shape = {'label': label,
                 'score': sum(d['probability']) / len(d['probability']),
                 'points': [zoom_point2(pt, 1 / ratio) for pt in d['vertexes_location']],
                 'shape_type': 'polygon'}
        result_dict['shapes'] = [shape]

        if 'words_result' in result_dict:
            del result_dict['words_result']
        return result_dict

    def vinCode(self, image, **options):
        """ VIN码识别: https://cloud.baidu.com/doc/OCR/s/zk3h7y51e """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.vinCode, buffer, options)
        result_dict = ToLabelmeLike.list_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def vehicleInvoice(self, image, **options):
        """ 机动车销售发票识别: https://cloud.baidu.com/doc/OCR/s/vk3h7y4tx """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.vehicleInvoice, buffer, options)
        result_dict = ToLabelmeLike.dict_str(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def vehicleCertificate(self, image, **options):
        """ 车辆合格证识别: https://cloud.baidu.com/doc/OCR/s/yk3h7y3sc """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.vehicleCertificate, buffer, options)
        result_dict = ToLabelmeLike.dict_str(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def raw_mixed_multi_vehicle(self, image, **options):
        """ 车辆证照混贴识别: https://cloud.baidu.com/doc/OCR/s/Kksfsbngb """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.mixed_multi_vehicle, buffer, options)

        if ratio != 1:
            for x in result_dict['words_result']:
                x['location'] = {k: round_int(v / ratio) for k, v in x['location'].items()}

        return result_dict

    def mixed_multi_vehicle(self, image, **options):
        result_dict = self.raw_mixed_multi_vehicle(image, options)

        shapes = []
        for x in result_dict['words_result']:
            shape = {'label': {}}
            shape.update(loc2points(x['location']))
            shape['label']['text'] = json.dumps({w['word_name']: w['word'] for w in x['license_info']},
                                                ensure_ascii=False)
            shape['label']['category'] = x['card_type']
            shape['label']['score'] = x['probability']
            shapes.append(shape)

        result_dict['shapes'] = shapes
        del result_dict['words_result']
        del result_dict['words_result_num']
        return result_dict

    def vehicle_registration_certificate(self, image, **options):
        """ 机动车登记证书识别: https://cloud.baidu.com/doc/OCR/s/qknzs5zzo """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.vehicle_registration_certificate, buffer, options)
        result_dict = ToLabelmeLike.dict_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def weightNote(self, image, **options):
        """ 磅单识别: https://cloud.baidu.com/doc/OCR/s/Uksfp9far """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.weightNote, buffer, options)

        shapes = []
        for x in result_dict['words_result']:
            for k, vs in x.items():
                shape = {'category': k,
                         'label': ''.join([v['word'] for v in vs])}
                shapes.append(shape)

        result_dict['shapes'] = shapes
        del result_dict['words_result']
        del result_dict['words_result_num']
        return result_dict

    def __B4_财务(self):
        pass

    def multipleInvoice(self, image, **options):
        """ 智能财务票据识别: https://cloud.baidu.com/doc/OCR/s/7ktb8md0j """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.multipleInvoice, buffer, options)

        shapes = []
        for x in result_dict['words_result']:
            shape = {'category': x['type'], 'label': {'text': ''}}
            shape.update(loc2points(x, 1 / ratio))
            shapes.append(shape)

        result_dict['shapes'] = shapes
        del result_dict['words_result']
        del result_dict['words_result_num']
        return result_dict

    def quotaInvoice(self, image, **options):
        """ 定额发票识别: https://cloud.baidu.com/doc/OCR/s/lk3h7y4ev """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.quotaInvoice, buffer, options)
        result_dict = ToLabelmeLike.dict_str(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def invoice(self, image, **options):
        """ 通用机打发票识别: https://cloud.baidu.com/doc/OCR/s/Pk3h7y06q """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.invoice, buffer, options)

        shapes = []
        for k, v in result_dict['words_result'].items():
            shape = {'label': {'category': k}}
            if isinstance(v, list):
                shape['label']['text'] = '\n'.join([w['word'] for w in v])
            else:
                shape['label']['text'] = v
                shapes.append(shape)

        result_dict['shapes'] = shapes
        del result_dict['words_result']
        del result_dict['words_result_num']
        return result_dict

    def trainTicket(self, image, **options):
        """ 火车票识别: https://cloud.baidu.com/doc/OCR/s/Ok3h7y35u """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.trainTicket, buffer, options)
        result_dict = ToLabelmeLike.dict_str(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def taxiReceipt(self, image, **options):
        """ 出租车票识别: https://cloud.baidu.com/doc/OCR/s/Zk3h7xxnn """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.taxiReceipt, buffer, options)
        result_dict = ToLabelmeLike.dict_str(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def airTicket(self, image, **options):
        """ 飞机行程单识别: https://cloud.baidu.com/doc/OCR/s/Qk3h7xzro """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.airTicket, buffer, options)
        result_dict = ToLabelmeLike.dict_str(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def onlineTaxiItinerary(self, image, **options):
        """ 网约车行程单识别: https://cloud.baidu.com/doc/OCR/s/Bkocoyu9n """

        def func(image, options=None):
            return self._aipocr.onlineTaxiItinerary(image)

        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(func, buffer, options, mode_name='onlineTaxiItinerary')

        shapes = []
        for k, v in result_dict['words_result'].items():
            if k == 'items':
                shape = {'label': {'category': 'item'}}
                shape['label']['text'] = json.dumps(v, ensure_ascii=False)
            else:
                shape = {'label': {'category': k}}
                shape['label']['text'] = v
            shapes.append(shape)

        result_dict['shapes'] = shapes
        del result_dict['words_result']
        del result_dict['words_result_num']
        return result_dict

    def receipt(self, image, **options):
        """ receipt: https://cloud.baidu.com/doc/OCR/s/6k3h7y11b """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.receipt, buffer, options)
        result_dict = ToLabelmeLike.list_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def __B5_医疗(self):
        pass

    def raw_medicalInvoice(self, image, **options):
        """ 医疗发票识别: https://cloud.baidu.com/doc/OCR/s/yke30j1hq """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.medicalInvoice, buffer, options)
        return result_dict

    def raw_medicalDetail(self, image, **options):
        """ 医疗费用明细识别: https://cloud.baidu.com/doc/OCR/s/Bknjnwlyj """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.medicalDetail, buffer, options)
        return result_dict

    def raw_insuranceDocuments(self, image, **options):
        """ 保险单识别: https://cloud.baidu.com/doc/OCR/s/Wk3h7y0eb """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.insuranceDocuments, buffer, options)
        return result_dict

    def __B6_教育(self):
        pass

    def docAnalysis(self, image, **options):
        """ 试卷分析与识别: https://cloud.baidu.com/doc/OCR/s/jk9m7mj1l
        总量1000次
        """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.docAnalysis, buffer, options)

        shapes = []
        for x in result_dict['results']:
            shape = loc2points(x['words']['words_location'], 1 / ratio)
            # 有line_probability字段，但实际并没有返回置信度~
            shape['label'] = {'category': x['words_type'], 'text': x['words']['word']}
            shapes.append(shape)

        result_dict['shapes'] = shapes
        del result_dict['results']
        del result_dict['results_num']
        return result_dict

    def formula(self, image, **options):
        """ 公式识别: https://cloud.baidu.com/doc/OCR/s/Ok3h7xxva
        总共1000次
        """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.formula, buffer, options)
        result_dict = ToLabelmeLike.list_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def __B7_其他(self):
        pass

    def meter(self, image, **options):
        """ 仪器仪表盘读数识别: https://cloud.baidu.com/doc/OCR/s/Jkafike0v """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.meter, buffer, options)
        result_dict = ToLabelmeLike.list_word(result_dict, 1 / ratio, 'words_result', 'words_result_num')
        return result_dict

    def raw_facade(self, image, **options):
        """ 门脸文字识别: https://cloud.baidu.com/doc/OCR/s/wk5hw3cvo """
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(self._aipocr.facade, buffer, options)
        return result_dict

    def facade(self, image, **options):
        result_dict = self.raw_facade(image, options)

        shapes = []
        for x in result_dict['words_result']:
            shape = {'label': {'text': x['words'], 'score': x['score']}}
            shapes.append(shape)

        result_dict['shapes'] = shapes
        del result_dict['words_result']
        del result_dict['words_result_num']
        return result_dict

    def __C_其他三方api接口(self):
        pass

    def mathpix_latex(self, image, **options):
        """ 调用mathpix识别单张图片的公式

        【return】
        {'auto_rotate_confidence': 0.0003554748584377876,
         'auto_rotate_degrees': 0,
         'detection_list': ['is_printed'],
         'detection_map': {'contains_chart': 0,
                           'contains_diagram': 0,
                           'contains_graph': 0,
                           'contains_table': 0,
                           'is_blank': 0,
                           'is_inverted': 0,
                           'is_not_math': 0,
                           'is_printed': 0.9996553659439087},
         'error': '',
         'latex': '\\left. \\begin{array} { l } { \\alpha / / \\beta } \\\\ { \\gamma '
                  '\\cap \\alpha = a } \\\\ { \\gamma \\cap \\beta = b } \\end{array} '
                  '\\right\\} \\Rightarrow a / / b',
         'latex_confidence': 0.9824940343387425,
         'latex_confidence_rate': 0.9994673295454546,
         'latex_list': [],
         'position': {'height': 160, 'top_left_x': 0, 'top_left_y': 0, 'width': 266},
         'request_id': '2022_05_28_762803ab687ff5ba1d80g'}
        """

        def func(buffer, options):
            image_uri = f'data:image/jpg;base64,' + base64.b64encode(buffer).decode()
            for i in range(3):  # 尝试3次
                try:
                    r = requests.post('https://api.mathpix.com/v3/latex',
                                      data=json.dumps({'src': image_uri}),
                                      headers=self._mathpix_header)
                    return json.loads(r.text)
                except requests.exceptions.ConnectionError:
                    pass
            else:
                return {'latex': '', 'error': 'requests.exceptions.ConnectionError'}

        # mathpix的接口没有说限制图片大小，但我还是按照百度的规范处理下更好
        buffer, ratio = self.adjust_image(image)
        result_dict = self.run_aipocr_with_db(func, buffer, options, mode_name='mathpix_latex')
        if 'position' in result_dict:
            result_dict['position'] = zoom_point(result_dict['position'], ratio)
        return result_dict

    def __D_福建省模式识别与图像理解重点实验室(self):
        pass

    def _priu_read_image(self, image, min_length=None, max_length=None,
                         limit_b64buffer_size=None, b64encode=None, **kwargs):
        # 默认不进行尺寸、文件大小压缩，这样会由平台上负责进行ratio缩放计算
        buffer, ratio = self.adjust_image(image,
                                          min_length=min_length, max_length=max_length,
                                          limit_b64buffer_size=limit_b64buffer_size,
                                          b64encode=True, **kwargs)
        assert ratio == 1, f'本地不做缩放，由服务器进行缩放处理'
        return buffer.decode()

    def priu_api(self, mode, image=None, texts=None, options=None, timeout=30, **meta_opts):
        """ 借助服务器来调用该类中含有的其他函数接口

        使用该接口的时候，因为服务器一般会有图片等的备份，所以本接口默认不对图片进行备份
        另外因为可能会调用自定义的模型功能，自定义的模型可能迭代较快，不适合在数据库缓存结果，所以也不记录json结果

        :param mode: 使用的api接口名称
        :param image: image、texts、options都是比较常用的几个参数键值，所以显式地写出来
        :param options: 有时候为了严谨，api相关的参数可能要单独打包
            但其实这里的参数混入meta_opts也没事，底层会有divide拆分函数，能区分开不同的参数类型
        :param timeout: 超时30秒的限制，有些特殊情况可以设置的更短
        :param meta_opts: 也可以自己额外扩展一些键值，兼容一些特殊的输入范式的api接口
            use_exists_xlapi: 是否复用已有的识别结果，避免重复运算，节约api成本
            record_files: 是否保存图片等数据
            record_xlapi: 是否保存api执行结果
            record_xlserver: 是否保存本次调用纪录
            根据不同的api，这些参数有不同的默认值，但一般都是True
        """
        # 1 统一输入参数的范式
        data = {}
        if image is not None:
            data['image'] = self._priu_read_image(image, b64encode=True)
        if texts is not None:
            data['texts'] = texts
        if options:  # 支持对算法的参数打包，但其实也支持从kwargs直接零散输入
            data['options'] = options
        if meta_opts:
            data.update(meta_opts)

        try:
            r = requests.post(f'{self._priu_host}/api/{mode}',
                              json=data, headers=self._priu_header, timeout=timeout)
        except requests.exceptions.Timeout:
            # 为了方便处理，统一也返回ConnectionError
            raise requests.exceptions.ConnectionError(r.text)

        if r.status_code == 200:
            res = json.loads(r.text)
        else:  # TODO 正常状态码不只200，可能还有重定向等某些不一定是错误的状态
            raise requests.exceptions.ConnectionError(r.text)

        # 2 统一返回值的范式，默认都是dict。 有些特殊格式表示是图片，这里会自动做后处理解析。
        if isinstance(res, dict) and len(res) == 1 and 'imageData' in res:
            # 只有一张图片情况的数据，直接返回图片
            return xlcv.read_from_buffer(res['imageData'], b64decode=True)
        else:
            return res

    def common_ocr(self, image, **options):
        """ 通用文字识别 """
        options.update({'image': self._priu_read_image(image, b64encode=True)})
        r = requests.post(f'{self._priu_host}/api/common_ocr', headers=self._priu_header,
                          json=options)
        res = json.loads(r.text)
        return res

    def ocr2texts(self, image, mode='common_ocr', **options):
        """ 通用的识别一张图的所有文本

        >> xlapi.ocr2texts(im)
        ['OCR文字识别软件', '如何将图片转换成Word', '文章来源：OCR文字识别软件', 'OCR']
        """
        texts = []  # 因图片太小等各种原因，没有识别到结果，默认就设空值
        try:
            d = self.priu_api(mode, image, options=options)
            if 'shapes' in d:
                texts = [sp['label']['text'] for sp in d['shapes']]
        except requests.exceptions.ConnectionError:
            pass
        return texts

    def rec_singleline(self, image, mode='common_ocr', **options):
        """ 通用的识别一张图的所有文本，并拼接到一起 """
        texts = self.ocr2texts(image, mode, **options)
        return ' '.join(texts)

    def hesuan_layout(self, image):
        """ 核酸版面分析 """
        r = requests.post(f'{self._priu_host}/api/hesuan_layout', headers=self._priu_header,
                          json={'image': self._priu_read_image(image)})
        return json.loads(r.text)

    def lexical_analysis(self, texts, options=None, return_mode=None):
        """ 词法分析，有分词的效果

        >> lexical_analysis(["今天是个好日子", "天气预报说今天要下雨"])
        [['今天', '是', '个', '好日子'], ['天气预报', '说', '今天', '要', '下雨']]
        """
        # TODO 可以增加接口参数，配置返回值类型。其他api同理。
        data = {'texts': texts}
        if options:
            data['options'] = options
        r = requests.post(f'{self._priu_host}/api/lac', json=data, headers=self._priu_header)
        res = json.loads(r.text)

        if return_mode == 'raw':
            pass
        else:
            res = [x['word'] for x in res]
        return res

    def sentiment_classify(self, texts, options=None):
        """ 情感分析

        返回的值越大表示句子的情感越积极
        >> sentiment_classify(['今天是个好日子', '天气预报说今天要下雨'])
        [0.4555, 0.2513]
        """
        data = {'texts': texts}
        if options:
            data['options'] = options
        r = requests.post(f'{self._priu_host}/api/senta_bilstm', json=data, headers=self._priu_header)
        return json.loads(r.text)

    def humanseg(self, image):
        """ 人像抠图

        TODO 还没测过特大、特小图会不会有问题~
        """
        im, _ = self.adjust_image(image, 1, to_buffer=False)  # 不能用xlcv.read，因为也可能作为服务端接口，输入buffer格式
        mask = self.priu_api('deeplabv3p_xception65_humanseg', im)
        new_im = np.concatenate([im, np.expand_dims(mask, axis=2)], axis=2)  # 变成BGRA格式图片
        return new_im

    def det_face(self, image):
        """ 人脸识别 """
        lmdict = self.priu_api('ultra_light_fast_generic_face_detector_1mb_640', image)
        return lmdict

    def super_resolution(self, image, times=1):
        """ 超分辨率（把图片变高清）

        :param times: 放大多少次，每次放大2倍（长宽个放大2倍）
        """
        for _ in range(times):
            image = self.priu_api('falsr_c', image)
        return image

    def rec_speech(self, audio_file):
        """ 语音识别 """
        if os.path.isfile(audio_file):
            audio = base64.b64encode(XlPath(audio_file).read_bytes()).decode()
        else:
            raise NotImplementedError
        text = self.priu_api('u2_conformer_wenetspeech', audio=audio)[0]
        return text


def demo_aipocr():
    import pprint
    import re

    from pyxlpr.data.labelme import LabelmeDict

    xlapi = XlAiClient()
    # xlapi.setup_database()
    xlapi._priu_host = 'http://localhost:5003'

    mode = 'general'

    _dir = XlPath("/home/chenkunze/data/aipocr_test")
    fmode = re.sub(r'^raw_', r'', mode)
    fmode = {'basicAccurate': 'accurate',
             'basicGeneral': 'general',
             'webimageLoc': 'webImage',
             'idcard_back': 'idcard',
             'vat_invoice_verification': 'vatInvoice',
             'mathpix_latex': 'formula',
             }.get(fmode, fmode)
    files = _dir.glob_images(f'*/{fmode}/**/*')

    for f in list(files):
        # 1 处理指定图片
        # if f.stem != '824ef2b9a5f05422199107721d299f30':
        #     continue

        # 2 检查字典
        print(f.as_posix())
        d = getattr(xlapi, mode)(f.as_posix())

        # browser.html(d['htmltables'][0])
        print()
        pprint.pprint(d)
        print('- ' * 20)

        # 3 前置的错误图可以删除；有shapes的可以转labelme；非labelme格式上面print后直接退出
        if 'error_code' in d:
            f.delete()
        elif d.get('shapes', 0):
            # tolabelme
            lmdata = LabelmeDict.gen_data(f)
            lmdata['shapes'] = d['shapes']
            for sp in lmdata['shapes']:
                sp['label'] = json.dumps(sp['label'], ensure_ascii=False)
            f.with_suffix('.json').write_json(lmdata)
            break
        else:
            break



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/06 18:40

from pyxllib.prog.specialist.common import *
from pyxllib.prog.specialist.xllog import *
from pyxllib.prog.specialist.browser import *
from pyxllib.prog.specialist.bc import *
from pyxllib.prog.specialist.tictoc import *
from pyxllib.prog.specialist.datetime import *

import concurrent.futures
import os
import re
import subprocess
import time
from statistics import mean
from threading import Thread

from tqdm import tqdm
import requests
from humanfriendly import parse_size

from pyxllib.prog.newbie import human_readable_size
from pyxllib.prog.pupil import get_installed_packages
from pyxllib.prog.xlosenv import XlOsEnv
from pyxllib.file.specialist import cache_file


def mtqdm(func, iterable, *args, max_workers=1, check_per_seconds=0.01, **kwargs):
    """ 对tqdm的封装，增加了多线程的支持

    这里名称前缀多出的m有multi的意思

    :param max_workers: 默认是单线程，改成None会自动变为多线程
        或者可以自己指定线程数
        注意，使用负数，可以用对等绝对值数据的“多进程”
    :param smoothing: tqdm官方默认值是0.3
        这里关掉指数移动平均，直接计算整体平均速度
        因为对我个人来说，大部分时候需要严谨地分析性能，得到整体平均速度，而不是预估当前速度
    :param mininterval: 官方默认值是0.1，表示显示更新间隔秒数
        这里不用那么频繁，每秒更新就行了~~
    :param check_per_seconds: 每隔多少秒检查队列
        有些任务，这个值故意设大一点，可以减少频繁的队列检查时间，提高运行速度
    整体功能类似Iterate
    """

    # 0 个人习惯参数
    kwargs['smoothing'] = kwargs.get('smoothing', 0)
    kwargs['mininterval'] = kwargs.get('mininterval', 1)

    if max_workers == 1:
        # 1 如果只用一个线程，则不使用concurrent.futures.ThreadPoolExecutor，能加速
        for x in tqdm(iterable, *args, **kwargs):
            func(x)
    else:
        # 2 默认的多线程运行机制，出错是不会暂停的；这里对原函数功能进行封装，增加报错功能
        error = False

        def wrap_func(x):
            nonlocal error
            try:
                func(x)
            except Exception as e:
                error = e

        # 3 多线程/多进程 和 进度条 功能的结合
        if max_workers > 1:
            executor = concurrent.futures.ThreadPoolExecutor(max_workers)
            for x in tqdm(iterable, *args, **kwargs):
                while executor._work_queue.qsize():
                    if check_per_seconds:
                        time.sleep(check_per_seconds)
                executor.submit(wrap_func, x)
                if error:
                    raise error
        else:
            executor = concurrent.futures.ProcessPoolExecutor(-max_workers)
            for x in tqdm(iterable, *args, **kwargs):
                # while executor._call_queue.pending_work_items:
                #     if check_per_seconds:
                #         time.sleep(check_per_seconds)
                executor.submit(wrap_func, x)
                if error:
                    raise error

        executor.shutdown()


def distribute_package(root, version=None, repository=None, *,
                       upload=True,
                       version_file='setup.py',
                       delete_dist=True):
    """ 发布包的工具函数

    :param root: 项目的根目录，例如 'D:/slns/pyxllib'
        根目录下有对应的 setup.py 等文件
    :param repository: 比如我配置了 [xlpr]，就可以传入 'xlpr'
    :param version_file: 保存版本号的文件，注意看正则规则，需要满足特定的范式，才会自动更新版本号
    :param delete_dist: 上传完是否自动删除dist目录，要检查上传包是否有遗漏时，要关闭
    """
    from pyxllib.file.specialist import XlPath

    # 1 切换工作目录
    os.chdir(str(root))

    # 2 改版本号
    if version:
        f = XlPath(root) / version_file
        s = re.sub(r"(version\s*=\s*)(['\"])(.+?)(\2)", fr'\1\g<2>{version}\4', f.read_text())
        f.write_text(s)

    # 3 打包
    subprocess.run('python setup.py sdist')

    # 4 上传
    if upload:
        # 上传
        cmd = 'twine upload dist/*'
        if repository:
            cmd += f' -r {repository}'
        subprocess.run(cmd)
        # 删除打包生成的中间文件
        if delete_dist:
            XlPath('dist').delete()
        XlPath('build').delete()

        # 这个不能删，不然importlib会读取不到模块的版本号
        # [d.delete() for d in XlPath('.').select_dirs(r'*.egg-info')]


def estimate_package_size(package):
    """ 估计一个库占用的存储大小 """

    # 将cache文件存储到临时目录中，避免重复获取网页
    def get_size(package):
        r = requests.get(f'https://pypi.org/project/{package}/#files')
        if r.status_code == 404:
            return '(0 MB'  # 找不到的包默认按0MB计算
        else:
            return r.text

    s = cache_file(package + '.pypi', lambda: get_size(package))
    # 找出所有包大小，计算平均值作为这个包大小的预估
    # 注意，这里进位是x1000，不是x1024
    v = mean(list(map(parse_size, re.findall(r'\((\d+(?:\.\d+)?\s*\wB(?:ytes)?)', s))) or [0])
    return v


def estimate_pip_packages(*, print_mode=False):
    """ 检查pip list中包的大小，从大到小排序

    :param print_mode:
        0，不输出，只返回运算结果，[(package_name, package_size), ...]
        1，输出最后的美化过的运算表格
        2，输出中间计算过程
    """

    def printf(*args, **kwargs):
        # dm表示mode增量
        if print_mode > 1:
            print(*args, **kwargs)

    packages = get_installed_packages()
    package_sizes = []
    for package_name in packages:
        package_size = estimate_package_size(package_name)
        package_sizes.append((package_name, package_size))
        printf(f"{package_name}: {human_readable_size(package_size)}")

    package_sizes.sort(key=lambda x: (-x[1], x[0]))
    if print_mode > 0:
        if print_mode > 1: print('- ' * 20)
        for package_name, package_size in package_sizes:
            print(f"{package_name}: {human_readable_size(package_size)}")
    return package_sizes


class ProgressBar:
    """ 对运行可能需要较长时间的任务，添加进度条显示

    # 示例用法
    with ProgressBar(100) as pb:
        for i in range(100):
            time.sleep(0.1)  # 模拟耗时工作
            pb.progress = i + 1  # 更新进度
    """
    def __init__(self, total):
        self.total = total  # 总进度
        self.progress = 0  # 当前进度
        self.stop_flag = False  # 停止标志

    def __enter__(self):
        # 启动进度显示线程
        self.progress_thread = Thread(target=self.display_progress)
        self.progress_thread.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # 强制将进度设置为100%
        self.progress = self.total
        # 停止进度显示线程
        self.stop_flag = True
        self.progress_thread.join()

    def display_progress(self):
        with tqdm(total=self.total) as pbar:
            while not self.stop_flag:
                pbar.n = self.progress
                pbar.refresh()
                time.sleep(1)
            pbar.n = self.progress
            pbar.refresh()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2020/09/15 21:00


from pyxllib.xlcv import *


class RemoveLine:
    def __init__(self, filename):
        self.bgr_img = xlcv.read(filename)
        self.gray_img = cv2.cvtColor(self.bgr_img, cv2.COLOR_BGR2GRAY)
        self.binary_img = cv2.adaptiveThreshold(self.gray_img, 255, 0, 1, 11, 3)

    def debug(self):
        """ 同run，只是debug模式，会输出中间结果
        """
        bgr_img = self.bgr_img
        xlcv.show(bgr_img, '0 src')
        dprint(bgr_img.shape)

        lines = self.detect_lines()
        xlcv.show(xlcv.lines(bgr_img, lines, [0, 0, 255]), '1 hough')
        lines = self.refine_lines(lines)
        xlcv.show(xlcv.lines(bgr_img, lines, [0, 0, 255]), '2 expand')
        dprint(lines.shape)

        dst = self.remove_lines(lines)
        xlcv.show(dst, 'result')

    def run(self):
        """
        :return: 获得去掉线后的图片
        """
        lines = self.detect_lines()
        if lines.any():
            lines = self.refine_lines2(lines)
            return self.remove_lines(lines)
        else:
            return self.bgr_img

    def detect_lines(self, rho=1, theta=np.pi / 180, threshold=80, min_line_length=50, max_line_gap=30):
        lines = cv2.HoughLinesP(self.binary_img, rho, theta, threshold, min_line_length, max_line_gap)
        # 不知道为什么返回值第2维会多一个1，把它删了~
        lines = np.array([]) if lines is None else lines.reshape(-1, 4)
        return lines

    def refine_lines(self, lines):
        im = self.binary_img
        n, m = im.shape

        def f(v):
            """ 辅助函数：四舍五入取整 """
            return int(round(v))

        def expand(x, y, dx, dy):
            """ 从(x,y)开始，按dx,dy逐步遍历，直到遇到im二值图为0的位置 """
            while True:
                j, i = f(x + dx), f(y + dy)
                if 0 <= j < m and 0 <= i < n and im[i][j]:
                    x, y = x + dx, y + dy
                else:
                    return f(x), f(y)

        new_lines = []
        for line in lines:
            # 1 确保x1是左边，x2是右边
            x1, y1, x2, y2 = line
            if x2 < x1:
                x1, y1, x2, y2 = x2, y2, x1, y1
            # 2 计算斜率，分两类处理
            if abs(x1 - x2) > abs(y1 - y2):
                dx, dy = 1, (y1 - y2) / (x1 - x2)
            else:
                dx, dy = (x1 - x2) / (y1 - y2), 1
            # 3 向左右各自延展
            x1, y1 = expand(x1, y1, -dx, -dy)
            x2, y2 = expand(x1, y1, dx, dy)
            new_lines.append((x1, y1, x2, y2))
        return np.array(new_lines)

    def refine_lines2(self, lines):
        """ 标准的延展会有些问题，会误删很多斜的笔划

        这个版本的删除，会只删除比较水平的情况，宁愿少删，但不误删
        """
        im = self.binary_img
        n, m = im.shape

        def f(v):
            """ 辅助函数：四舍五入取整 """
            return int(round(v))

        def expand(x, y, dx, dy):
            """ 从(x,y)开始，按dx,dy逐步遍历，直到遇到im二值图为0的位置 """
            while True:
                j, i = f(x + dx), f(y + dy)
                if 0 <= j < m and 0 <= i < n and im[i][j]:
                    x, y = x + dx, y + dy
                else:
                    return f(x), f(y)

        new_lines = []
        for line in lines:
            # 1 确保x1是左边，x2是右边
            x1, y1, x2, y2 = line
            if x2 < x1:
                x1, y1, x2, y2 = x2, y2, x1, y1
            # 2 计算斜率，分两类处理
            dx, dy = 1, (y1 - y2) / (x1 - x2 + sys.float_info.epsilon)
            if abs(dy) > math.tan(math.pi / 180 * 10):
                # 倾斜度必须在10度以内（偏水平的线）
                continue
            # 3 向左右各自延展
            x1, y1 = expand(x1, y1, -dx, -dy)
            x2, y2 = expand(x1, y1, dx, dy)
            new_lines.append((x1, y1, x2, y2))
        return np.array(new_lines)

    def remove_lines(self, lines, box_size=(7, 3), num=5):
        """
        :param lines:
        :param box_size: 所有尺寸，默认是7*3的矩阵
            7*3矩阵中，如果有超出4个前景点，则恢复该背景点为原始像素值
        :param num: 矩阵至少需要的点数量
        :return:
        """
        # 1 计算填充用的默认背景色
        src = self.bgr_img
        if not lines.any(): return src
        bg_color = xlcv.bg_color(src, binary_img=self.binary_img)
        # dprint(bg_color)

        # 2 删笔画的mask矩阵
        mask = xlcv.lines(np.zeros(src.shape[:2]), lines, [255], 2)
        tmp = xlcv.lines(src, lines, bg_color, 2)  # 在原图画上背景颜色的线，产生抹除效果
        dst = np.array(tmp)
        tmp = cv2.cvtColor(tmp, cv2.COLOR_BGR2GRAY)

        # 3 恢复断开的笔划
        top, left = box_size[0] // 2, box_size[1] // 2
        bottom, right = box_size[0] - top, box_size[1] - left
        for i in range(src.shape[0]):
            for j in range(src.shape[1]):
                if not mask[i, j]: continue
                arr = tmp[max(i - top, 0):i + bottom, max(j - left, 0):j + right].reshape(-1)
                if sum(arr < 128) >= num:
                    dst[i, j] = src[i, j]
        # xlcv.show(tmp)
        return dst


def test_removeline(path):
    """ 批量处理，检查去线功能效果 """

    def func(p1, p2):
        dst = RemoveLine(str(p1)).run()
        cv2.imwrite(str(p2), dst)

    d1, d2 = Dir(path), Dir(path + '+')
    d2.ensure_dir()
    d1_state = d1.select('*.png')
    d1_state.subs = d1_state.subs
    d1_state.procpaths(func, ref_dir=d2, pinterval=1000)


if __name__ == '__main__':
    TicToc.process_time(f'{dformat()}启动准备共用时')
    tictoc = TicToc(__file__)
    os.chdir(r'D:\RealEstate2020')

    # RemoveLine(r'5_label+\agreement_label1_text2\handwriting\C100001245519-003.png').debug()
    test_removeline(r'5_label+\agreement_label1_text2\handwriting')
    test_removeline(r'5_label+\agreement_label1_text2\printed')  # 65680张图，704秒

    # cv2.waitKey(0)
    tictoc.toc()
